#include <stdio.h>
#include "mpi.h"
#include "comm.h"
#include "globals.h"
#include "utils.h"



void onREvent( RequestTag* );
void onWEvent( RequestTag* );
void onSEvent( RequestTag* );
static EventHandler eventHandlers[5] = 
   { &onREvent, 
     &onWEvent,
     &onWEvent,
     &onSEvent,
     &onSEvent
   };


/**
 * Macros for resumable functions
 */
#define BEGINRESUMABLE \
   int _thiscnt = 0; 
 
#define BEGINTASK \
   if( _thiscnt++ >= *cnt ) { \
      *cnt = _thiscnt;

#define ENDTASK \
      return 1; \
   }

#define BEGINPSEUDOTASK \
   if( _thiscnt++ >= *cnt ) { \
      *cnt = _thiscnt;

#define ENDPSEUDOTASK \
   }



/**
 * Communication MPI tags
 */
/* R: 0 */
#define COMM_TAG_R  0 
/* W1: (1) upto (1 + rblks) */
#define COMM_TAG_W1 1
/* W2: (1 + rblks) upto (1 + 2*rblks) */
#define COMM_TAG_W2 (COMM_TAG_W1 + globals.A->rblks)
#define COMM_TAG_W0(wbuf) ((wbuf) == 0 ? COMM_TAG_W1 : COMM_TAG_W2)
/* S1: (1) upto (1 + cblks) */
#define COMM_TAG_S1 1
/* S2: (1 + cblks) upto (1 + 2*cblks) */
#define COMM_TAG_S2 (COMM_TAG_S1 + globals.A->cblks)
#define COMM_TAG_S0(sbuf) ((sbuf) == 0 ? COMM_TAG_S1 : COMM_TAG_S2)
   
   

/**
 * The resumable functions (multistage communication) interfaced by the comm subsystem 
 */
int communicateR( int j, int *cnt );
int communicateW( int j, int i, int *cnt );
int communicateS( int j, int k, int *cnt );

/**
 * Called as a response from the comm subsystem when an MPI request finishes.
 * These functions trigger the communicate* functions above.
 */
void onREvent( RequestTag *rt );
void onWEvent( RequestTag *rt );
void onSEvent( RequestTag *rt );

/**
 * Start the communication of a specific buffer block. 
 * These functions make sure that any previous communication on the block finishes first.
 * Triggers the communicate_ functions above with *cnt = 0.
 */
void startCommunicateR( int j );
void startCommunicateW( int j, int i );
void startCommunicateS( int j, int k );


/**
 * Allocates a BlockVectorBuffer consisting of 'num' blocks of size 'blksz' elements each.
 * \param[in] num The number of blocks in the BlockVectorBuffer.
 * \param[in] blksz The number of elements per block.
 * \return An allocated and initialized BlockVectorBuffer.
 */
BlockVectorBuffer *allocate_block_vector_buffer( int num, int blksz )
{
   BlockVectorBuffer *bvb;
   int i;

   bvb = abortmalloc( sizeof(BlockVectorBuffer) );
   bvb->matrix = abortmalloc( sizeof(double) * blksz * num );
   bvb->bbs = abortmalloc( sizeof(BufferBlockState) * num );
   for( i = 0; i < num; i++ ) {
      bvb->bbs[i].cnt = 0;
      bvb->bbs[i].is_executing = 0;
      bvb->bbs[i].state = -1;
   }

   return bvb;
}


/**
 * Deallocates a BlockVectorBuffer.
 * \param[in,out] bvb The BlockVectorBuffer to deallocate.
 */
void deallocate_block_vector_buffer( BlockVectorBuffer *bvb )
{
   free( bvb->bbs );
   free( bvb->matrix );
   bvb->bbs = NULL;
   bvb->matrix = NULL;
   free( bvb );
}


/**
 * Initializes the communication subsystem (stored at globals.cd).
 * This includes allocating arrays for MPI_Request:s, tags, statuses, indices, and communicators.
 * \param[in] capacity Hard upper limit on the number of concurrent requests.
 * \param[in] nprow The number of processor rows in the mesh.
 * \param[in] npcol The number of processor columns in the mesh.
 * \param[in] myrow Caller's process row.
 * \param[in] mycol Caller's process column.
 */
void comm_init( int capacity, int nprow, int npcol, int myrow, int mycol )
{
   /* Allocation of memory */
   globals.cd.mpi_reqs             = (MPI_Request*) abortmalloc( capacity * sizeof(MPI_Request) );
   globals.cd.tags                 = (RequestTag*) abortmalloc( capacity * sizeof(RequestTag) );
   globals.cd.statuses             = (MPI_Status*) abortmalloc( capacity * sizeof(MPI_Status) );
   globals.cd.indices              = (int*) abortmalloc( capacity * sizeof(int) );

   /* Initialization of numeric variables */
   globals.cd.capacity             = capacity;
   globals.cd.size                 = 0;

   /* Initialization of communicators */
   MPI_Comm_dup( MPI_COMM_WORLD, &globals.cd.comm_all );
   MPI_Comm_split( globals.cd.comm_all, mycol, myrow, &globals.cd.comm_col );
   MPI_Comm_split( globals.cd.comm_all, myrow, mycol, &globals.cd.comm_row );
}


/**
 * Frees memory allocated for communication subsystem.
 */
void comm_finalize( void )
{
   free( globals.cd.mpi_reqs );
   globals.cd.mpi_reqs = NULL;

   free( globals.cd.tags );
   globals.cd.tags = NULL;

   free( globals.cd.statuses );
   globals.cd.statuses = NULL;

   free( globals.cd.indices );
   globals.cd.indices = NULL;

   MPI_Comm_free( &globals.cd.comm_all );
   MPI_Comm_free( &globals.cd.comm_col );
   MPI_Comm_free( &globals.cd.comm_row );
}


/**
 * Adds an MPI_Request to the central datastructure for later polling.
 * \param[in] req The MPI_Request to add.
 * \param[in] tag The MPI tag used with 'req'.
 * \param[in] buf_ID The buffer ID.
 */
void comm_add_request( MPI_Request req, int tag, int buf_ID )
{
   int idx;

   if( req == MPI_REQUEST_NULL ) {
      fprintf( stderr, "!!! Request is NULL - will cause severe errors\n" );
      fflush( stderr );
      MPI_Abort( MPI_COMM_WORLD, 3 );
   }
   
   if( globals.cd.size < globals.cd.capacity ) {
      /* Add it at size */
      idx = globals.cd.size++;
      globals.cd.mpi_reqs[idx] = req;
      globals.cd.tags[idx].tag = tag;
      globals.cd.tags[idx].buf_ID = buf_ID;
   } else {
      fprintf( stderr, "!!! Too many requests added\n" );
      fflush( stderr );
      MPI_Abort( MPI_COMM_WORLD, 2 );
   }
}


/**
 * Poll (either blocking or non-blocking) all concurrent MPI_Requests.
 * Resume the algorithm associated with each completed request.
 * \param[in] flag_blocking 0: blocking (MPI_Wait), 1: non-blocking (MPI_Test).
 */
void comm_poll( int flag_blocking )
{
   int count, idx, i;

   /* Check for quick return if no pending requests. */
   if( globals.cd.size == 0 ) {
      return;
   }

   /* Test some */
   if( flag_blocking ) {
      MPI_Waitsome( globals.cd.size, globals.cd.mpi_reqs, &count, globals.cd.indices, globals.cd.statuses );
   } else {
      MPI_Testsome( globals.cd.size, globals.cd.mpi_reqs, &count, globals.cd.indices, globals.cd.statuses );
   }

   /* Resume all algorithms associated with the completed requests. */
   for( i = 0; i < count; i++ ) {
      idx = globals.cd.indices[i];
      /* Continue algorithm (NOTE: May add to the end of the lists) */
      (*eventHandlers[globals.cd.tags[idx].buf_ID])( &globals.cd.tags[idx] );
   }

   /* Pack lists */
   for( i = count-1; i >= 0; i-- ) {
      /* Reduce size by 1 */
      globals.cd.size--;
      /* Pack if necessary */
      idx = globals.cd.indices[i];
      if( idx < globals.cd.size ) {
         /* Move from globals.cd.size to idx */
         globals.cd.mpi_reqs[idx] = globals.cd.mpi_reqs[globals.cd.size];
         globals.cd.tags[idx] = globals.cd.tags[globals.cd.size];
         globals.cd.mpi_reqs[globals.cd.size] = MPI_REQUEST_NULL;
      } 
   }
}


/**
 * Wait until all algorithms have finished.
 */
void comm_wait_all( void )
{
   while( globals.cd.size > 0 )
      comm_poll( 1 );
}


/**
 * Wait until the BufferBlockState of 'bbs' reaches 'state'.
 * \param[in,out] bbs The BufferBlockState to wait on.
 * \param[in] The state value to wait for.
 */
void comm_wait( BufferBlockState *bbs, int state )
{
   while( bbs->state != state || bbs->is_executing )
      comm_poll( 1 );
}


/**
 * Wait until execution on the BufferBlockState finishes.
 * \param[in,out] bbs The BufferBlockState to wait on.
 */
void comm_wait_idle( BufferBlockState *bbs )
{
   while( bbs->is_executing )
      comm_poll( 1 );
}


/**
 * Initialize a BufferBlockState.
 * \param[out] bbs The BufferBlockState to initialize.
 */
void bbs_init( BufferBlockState *bbs )
{
   bbs->cnt    = 0;
   bbs->is_executing  = 0;
   bbs->state  = -1;
}


/**
 * ----------------------------------------------------------------------
 * Communication algorithms - returns 1 if they are not finished, 0 if they are finished
 */

/**
 * Replicate the pivot block (R buffer) on process column.
 * Only includes process rows that require this block.
 * \param[in] j The iteration number (0 to Nb-1).
 * \param[in,out] cnt Task counter used to pause and resume the algorithm.
 * \return 0: finished, 1: paused.
 */
int communicateR( int j, int *cnt )
{
   int pj;
   int up, down;
   int jl;
   int nprow, npcol, myrow, mycol;
   int Nb, blksz;
   MPI_Request req = MPI_REQUEST_NULL;

   BEGINRESUMABLE;

   nprow = globals.nprow;
   npcol = globals.npcol;
   myrow = globals.myrow;
   mycol = globals.mycol;
   Nb    = globals.A->Nb;
   blksz = globals.A->blksz;

   /* Proc(pj,mycol) holds A(j,j) */
   pj = (j % nprow);

   /* A(*,j) is local block column jl */
   jl = G2Lmin(mycol, npcol, j);
   
   /* The up/down neighbours */
   up    = (myrow + nprow - 1) % nprow;
   down  = (myrow + 1) % nprow;

   if( myrow != pj ) {
      /* I am not root (receive) */
      BEGINTASK {
         MPI_Irecv( &globals.R->matrix[0], blksz, MPI_DOUBLE, up, COMM_TAG_R, globals.cd.comm_col, &req );
         comm_add_request( req, COMM_TAG_R, BUF_ID_R );
      } ENDTASK 
   }

   BEGINPSEUDOTASK {
      globals.R->bbs[0].state = j;
   } ENDPSEUDOTASK

   if( down != pj && G2Gmin( down, nprow, j+1 ) < Nb ) {
      /* Neighbour down needs block (send) */
      BEGINTASK {
         MPI_Isend( &globals.R->matrix[0], blksz, MPI_DOUBLE, down, COMM_TAG_R, globals.cd.comm_col, &req );
         comm_add_request( req, COMM_TAG_R, BUF_ID_R );
      } ENDTASK
   }

   return 0;
}


/**
 * Replicates the W_i block from iteration j on process row.
 * Possibly splits into replication of the S_k block.
 * \param[in] j The iteration counter (0 to Nb-1)
 * \param[in] i The local block index of W.
 * \param[in,out] cnt Task counter used to pause and resume the algorithm.
 * \return 0: finished, 1: paused.
 */
int communicateW( int j, int i, int *cnt )
{
   int pi, qj, qk;
   int right, left;
   int ig, kl;
   int wbuf;
   int nprow, npcol, myrow, mycol;
   int Nb, blksz;
   MPI_Request req = MPI_REQUEST_NULL;

   BEGINRESUMABLE;

   nprow = globals.nprow;
   npcol = globals.npcol;
   myrow = globals.myrow;
   mycol = globals.mycol;
   Nb    = globals.A->Nb;
   blksz = globals.A->blksz;

   /* Find global i */
   ig = myrow + i*nprow;

   /* Process (pi, qj) holds W(i) for j */
   pi = ig % nprow;
   qj = j % npcol;
   
   /* Process (pi, qk) holds S(k) for j */
   qk = ig % npcol;
   
   /* The W buffer */
   wbuf = j % 2;

   /* Process neighbours */
   right = (mycol + 1) % npcol;
   left = (mycol + npcol - 1) % npcol;

   if( mycol != qj ) {
      /* I am not root (receive) */
      BEGINTASK {
         MPI_Irecv( &globals.W[wbuf]->matrix[i*blksz], blksz, MPI_DOUBLE, left, COMM_TAG_W0(wbuf) + i, globals.cd.comm_row, &req );
         comm_add_request( req, COMM_TAG_W0(wbuf) + i, BUF_ID_W1 + wbuf );
      } ENDTASK
   }

   BEGINPSEUDOTASK {
      globals.W[wbuf]->bbs[i].state = j;
   } ENDPSEUDOTASK

   if( mycol == qk ) {
      BEGINPSEUDOTASK {
         /* A(ig, ig) belongs to Proc(*,qk) and the local block column is kl */
         kl = G2Lmin( mycol, npcol, ig );
   
         /* Start S-communication on local column kl for iteration j */
         startCommunicateS( j, kl );
      } ENDPSEUDOTASK
   }

   if( right != qj && G2Gmin( right, npcol, j ) <= ig ) {
      BEGINTASK {
         MPI_Isend( &globals.W[wbuf]->matrix[i*blksz], blksz, MPI_DOUBLE, right, COMM_TAG_W0(wbuf) + i, globals.cd.comm_row, &req );
         comm_add_request( req, COMM_TAG_W0(wbuf) + i, BUF_ID_W1 + wbuf );
      } ENDTASK
   }
   
   return 0;
}


/**
 * Replicates the S_k block from iteration j on process column.
 * \param[in] j The iteration counter (0 to Nb-1)
 * \param[in] k The local block index of S.
 * \param[in,out] cnt Task counter used to pause and resume the algorithm.
 * \return 0: finished, 1: paused.
 */
int communicateS( int j, int k, int *cnt )
{
   int kg, il;
   int pk;
   int up, down;
   int sbuf;
   int nprow, npcol, myrow, mycol;
   int Nb, blksz;
   MPI_Request req = MPI_REQUEST_NULL;

   BEGINRESUMABLE;

   nprow = globals.nprow;
   npcol = globals.npcol;
   myrow = globals.myrow;
   mycol = globals.mycol;
   Nb    = globals.A->Nb;
   blksz = globals.A->blksz;

   /* Find global k */
   kg = mycol + k*npcol;

   /* Process (pk,mycol) holds S(k) for j */
   pk = kg % nprow;

   /* Neighbours */
   up = (myrow + nprow - 1) % nprow;
   down = (myrow + 1) % nprow;

   /* Local row of global row k */
   il = kg / nprow;

   /* The S buffer used */
   sbuf = j % 2;
   
   if( myrow == pk ) {
      /* I am root */
      BEGINPSEUDOTASK {
         copyblock( &globals.S[sbuf]->matrix[k*blksz], &globals.W[sbuf]->matrix[il*blksz], blksz );
      } ENDPSEUDOTASK 
   } else {
      /* I am not root */
      BEGINTASK {
         MPI_Irecv( &globals.S[sbuf]->matrix[k*blksz], blksz, MPI_DOUBLE, up, COMM_TAG_S0(sbuf) + k, globals.cd.comm_col, &req );
         comm_add_request( req, COMM_TAG_S0(sbuf) + k, BUF_ID_S1 + sbuf );                       
      } ENDTASK
   }
      
   BEGINPSEUDOTASK {
      globals.S[sbuf]->bbs[k].state = j;
   } ENDPSEUDOTASK 

   if( down != pk && G2Gmin( down, nprow, kg ) < Nb ) {
      BEGINTASK {
         MPI_Isend( &globals.S[sbuf]->matrix[k*blksz], blksz, MPI_DOUBLE, down, COMM_TAG_S0(sbuf) + k, globals.cd.comm_col, &req );
         comm_add_request( req, COMM_TAG_S0(sbuf) + k, BUF_ID_S1 + sbuf );                       
      } ENDTASK
   }

   return 0;
}



/**
 * Request completion callback for events associated with the R buffer.
 * \param[in] rt The RequestTag.
 */
void onREvent( RequestTag *rt )
{
   BufferBlockState *bbs;

   bbs = &globals.R->bbs[0];
   if( !communicateR( bbs->params.R.j, &bbs->cnt ) ) {
      /* Finished execution */
      assert(bbs->is_executing);
      bbs->is_executing = 0;
   }
}


/**
 * Request completion callback for events associated with the W buffers.
 * \param[in] rt The RequestTag.
 */
void onWEvent( RequestTag *rt )
{
   BufferBlockState *bbs;
   int widx;
   int wbuf;

   /* Determine from the buf_ID which one of W1 and W2 this request is for */
   wbuf = (rt->buf_ID == BUF_ID_W1 ? 0 : 1);

   /* Determine which block index into W this is for */
   widx = rt->tag - (rt->tag >= COMM_TAG_W2 ? COMM_TAG_W2 : COMM_TAG_W1);

   /* Continue the communication algorithm */
   bbs = &globals.W[wbuf]->bbs[widx];
   if( !communicateW( bbs->params.W.j, bbs->params.W.i, &bbs->cnt ) ) {
      /* Finished execution */
      assert(bbs->is_executing);
      bbs->is_executing = 0;
   }
}


/**
 * Request completion callback for events associated with the S buffers.
 * \param[in] rt The RequestTag.
 */
void onSEvent( RequestTag *rt )
{
   BufferBlockState *bbs;
   int sidx;
   int sbuf;

   /* Determine from the buf_ID which one of S1 and S2 this request is for */
   sbuf = (rt->buf_ID == BUF_ID_S1 ? 0 : 1);

   /* Determine which block index into W this is for */
   sidx = rt->tag - (rt->tag >= COMM_TAG_S2 ? COMM_TAG_S2 : COMM_TAG_S1);

   /* Continue the communication algorithm */
   bbs = &globals.S[sbuf]->bbs[sidx];
   if( !communicateS( bbs->params.S.j, bbs->params.S.k, &bbs->cnt ) ) {
      /* Finished execution */
      assert(bbs->is_executing);
      bbs->is_executing = 0;
   }
}


/**
 * Routines to start a communication algorithm (i.e. to store it in a BBS and run it once)
 * These routines make sure that no other algorithm is executing on the BBS.
 */

/**
 * Start to replicate R buffer for iteration j.
 * \param[in] j The iteration counter (0 to Nb-1).
 */
void startCommunicateR( int j )
{
   RequestTag rt;

   assert(!globals.R->bbs[0].is_executing);

   /* Reset BBS, store parameters, set execution flag */
   globals.R->bbs[0].cnt = 0;
   globals.R->bbs[0].params.R.j = j;
   globals.R->bbs[0].is_executing = 1;

   /* Perform one step of algorithm */
   rt.tag = COMM_TAG_R;
   rt.buf_ID = BUF_ID_R;
   onREvent( &rt );
}


/**
 * Start to replicate W_i for iteration j.
 * \param[in] j The iteration counter.
 * \param[in] i The block index.
 */
void startCommunicateW( int j, int i )
{
   int wbuf;
   RequestTag rt;

   /* Find the W buffer used */
   wbuf = j % 2;

   assert(!globals.W[wbuf]->bbs[i].is_executing);

   /* Reset BBS, store parameters, set execution flag */
   globals.W[wbuf]->bbs[i].cnt = 0;
   globals.W[wbuf]->bbs[i].params.W.j = j;
   globals.W[wbuf]->bbs[i].params.W.i = i;
   globals.W[wbuf]->bbs[i].is_executing = 1;

   /* Perform one step of algorithm */
   rt.buf_ID = BUF_ID_W1 + wbuf;
   rt.tag = COMM_TAG_W0(wbuf) + i;
   onWEvent( &rt );
}


/**
 * Start to replicate S_k for iteration j.
 * \param[in] j The iteration counter.
 * \param[in] k The block index.
 */
void startCommunicateS( int j, int k )
{
   int sbuf;
   RequestTag rt;

   /* Find the S buffer used */
   sbuf = j % 2;

   assert(!globals.S[sbuf]->bbs[k].is_executing);

   /* Reset BBS, store parameters, set execution flag */
   globals.S[sbuf]->bbs[k].cnt = 0;
   globals.S[sbuf]->bbs[k].params.S.j = j;
   globals.S[sbuf]->bbs[k].params.S.k = k;
   globals.S[sbuf]->bbs[k].is_executing = 1;

   /* Perform one step of algorithm */
   rt.buf_ID = BUF_ID_S1 + sbuf;
   rt.tag = COMM_TAG_S0(sbuf) + k;
   onSEvent( &rt );
}


