#include <stdio.h>
#include "dsbp.h"
#include "utils.h"


/**
 * Init a DSBP matrix by allocating storage, computing the cp array and initializing the struct fields.
 * \param[out] A The DSBPMatrix structure to initialize.
 * \param[in] nprow The number of process rows.
 * \param[in] npcol The number of process columns.
 * \param[in] myrow My process row.
 * \param[in] mycol My process column.
 * \param[in] N Matrix size.
 * \param[in] nb Block size.
 */
void DSBPInit( DSBPMatrix *A, int nprow, int npcol, int myrow, int mycol, int N, int nb )
{
   int Nb, Nbl, blksz;
   int cpsz;
   int j, jl, ilast;
   int sum;

   Nb = (N + nb - 1) / nb;
   blksz = nb * nb;

   A->N = N;
   A->nb = nb;
   A->Nb = Nb;
   A->blksz = blksz;
   
   /* Find number of local block columns (so we can allocate cp) */
   Nbl = Nb / npcol + (mycol < (Nb % npcol) ? 1 : 0);
   cpsz = Nbl;
   A->cp = abortmalloc( cpsz * sizeof(int) );

   /* Find the last global row */
   ilast = Nb - 1;

   /* Find the last local row index */
   A->ilastp = (ilast >= myrow ? (ilast - myrow) / nprow : -1);

   /* Loop through global columns that map to this process column */
   sum = 0;
   for( j = mycol, jl = 0; j < Nb; j += npcol, jl++ ) {
      sum += A->ilastp - (j - myrow + nprow - 1) / nprow + 1;
      A->cp[jl] = sum - 1;
   }

   /* Allocate data */
   A->matrix = abortmalloc( sum * blksz * sizeof(double) );

   /* Set size information */
   A->rblks = A->ilastp + 1;
   A->cblks = Nbl;
}


/**
 * Converts a distributed column major matrix A into DSBP format out-of-place by a trivial element-wise algorithm.
 * Used only for illustration/testing. 
 * \param[out] B The output DSBP matrix.
 * \param[in] A The input matrix.
 * \param[in] ldA The leading dimension of A.
 * \param[in] nprow The number of process rows.
 * \param[in] npcol The number of process columns.
 * \param[in] myrow My process row.
 * \param[in] mycol My process column.
 * \param[in] N The matrix size.
 * \param[in] nb The block size.
 */
void DSBPFromFull( DSBPMatrix *B, double *A, int ldA, int nprow, int npcol, int myrow, int mycol, int N, int nb )
{
   int is, ie, il, jl, j, ijl, r, c;
   int Nb, ib, jb, blksz;

   Nb = iceil( N, nb );
   blksz = nb*nb;

   DSBPInit( B, nprow, npcol, myrow, mycol, N, nb );

   ijl = 0;
   for( jl = 0; jl < B->cblks; jl++ ) {
      j = mycol + jl*npcol;
      is = G2Lmin( myrow, nprow, j );
      ie = B->ilastp;
      jb = min( nb, N - j*nb );
      for( il = is; il <= ie; il++, ijl++ ) {
         ib = min( nb, N - (myrow + il*nprow)*nb );
         for( c = 0; c < jb; c++ )  {
            for( r = 0; r < ib; r++ ) {
               B->matrix[ijl*blksz + r + c*nb] = A[il*nb+r + (jl*nb+c)*ldA];
            }
         }
      }
   }
}


/**
 * Outputs a DSBP matrix to the stream fp in MATLAB-readable form as a cell array of blocks.
 * The cell array is indexed globally so the output from each process can be merged trivially into a single MATLAB-readable form.
 * From the cell array the entire global matrix can be easily reconstructed by a double for loop.
 * \param[in,out] fp The output stream.
 * \param[in] A The DSBP matrix to output.
 * \param[in] nprow The number of process rows.
 * \param[in] npcol The number of process columns.
 * \param[in] myrow My process row.
 * \param[in] mycol My process column.
 */
void DSBP2Matlab( FILE *fp, DSBPMatrix *A, int nprow, int npcol, int myrow, int mycol )
{
   int il, jl, i, j, r, c, rb, cb, ijl;

   ijl = 0;
   for( jl = 0; jl < A->cblks; jl++ ) {
      j = mycol + jl*npcol;
      for( il = G2Lmin( myrow, nprow, j ); il <= A->ilastp; il++, ijl++ ) {
         i = myrow + il*nprow;
         fprintf(fp, "A{%d,%d} = [", i+1, j+1);
         rb = min(A->nb, A->N - i*A->nb);
         cb = min(A->nb, A->N - j*A->nb);
         for( r = 0; r < rb; r++ ) {
            for( c = 0; c < cb; c++ ) {
               fprintf(fp, "%.16e ", A->matrix[ijl*A->blksz + r + c*A->nb]);
            }
            fprintf(fp, "\n");
         }
         fprintf(fp, "];\n");
      }
   }
}


/**
 * Free up resources allocate for A.
 * \param[in,out] A The DSBP Matrix to clean up.
 */
void DSBPDestroy( DSBPMatrix *A )
{
   free( A->matrix );
   free( A->cp );
}


