#include "kernels.h"
#include "comm.h"
#include "utils.h"



int gINFO = 0;


/**
 * Extern declarations of FORTRAN BLAS/LAPACK routines used by the functions in this file.
 * C and FORTRAN "interoperability" is really platform specific.
 */
extern void dpotrf_(char*, int*, double*, int*, int*);
extern void dtrsm_(char*, char*, char*, char*, int*, int*, double*, double*, int*, double*, int*);
extern void dsyrk_(char*, char*, int*, int*, double*, double*, int*, double*, double*, int*);
extern void dgemm_(char*, char*, int*, int*, int*, double*, double*, int*, double*, int*, double*, double*, int*);


/**
 * Kernel POTRF.
 * Applies POTRF from iteration j to AL(ijl) which is A(iloc,jloc).
 * Makes sure the R-buffer is idle before copying factored block into it.
 * Sets the global gINFO to the first non-zero INFO it detects. 
 * This is used as a minimum safety net to detect if any error occured on some node during the computation.
 * \param[in] j The iteration counter.
 * \param[in] iloc The global row index of the block.
 * \param[in] jloc The global column index of the block.
 * \param[in] ijl The DSBP block index.
 */
void kernel_POTRF( int j, int iloc, int jloc, int ijl )
{
   int INFO;
   int N;

   N = min(globals.A->nb, globals.A->N - globals.A->nb * (globals.myrow + iloc*globals.nprow));
  
   if( opt_verbose ) {
      log_printf("POTRF(%d, %d, %d, %d)\n", j, iloc, jloc, ijl);
   }
   dpotrf_( "L", &N, &globals.A->matrix[globals.A->blksz * ijl], &globals.A->nb, &INFO );
   if( gINFO == 0 && INFO != 0 ) {
      gINFO = INFO; /* Save first non-zero INFO */
   }
   comm_wait_idle( &globals.R->bbs[0] );
   copyblock( &globals.R->matrix[0], &globals.A->matrix[ijl*globals.A->blksz], globals.A->blksz );
}


/**
 * Kernel TRSM.
 * Solves XR' = A for X, overwriting A (aka TRSM) from iteration j. Block A is AL(ijl) which is A(iloc,jloc).
 * Makes sure the W(iloc)-buffer block from set j%2 is idle before copying the scaled block into it.
 * \param[in] j The iteration counter.
 * \param[in] iloc The global row index of the A block.
 * \param[in] jloc The global column index of the A block.
 * \param[in] ijl The DSBP block index of the A block.
 */
void kernel_TRSM( int j, int iloc, int jloc, int ijl )
{
   double ONE = 1.0;

   int M, N;

   M = min(globals.A->nb, globals.A->N - globals.A->nb * (globals.myrow + iloc*globals.nprow));
   N = min(globals.A->nb, globals.A->N - globals.A->nb * (globals.mycol + jloc*globals.npcol));

   if( opt_verbose ) {
      log_printf("TRSM(%d, %d, %d, %d)\n", j, iloc, jloc, ijl);
   }
   dtrsm_( "R", "L", "T", "N", &M, &N, &ONE, &globals.R->matrix[0], &globals.A->nb, &globals.A->matrix[globals.A->blksz * ijl], &globals.A->nb );
   comm_wait_idle( &globals.W[j%2]->bbs[iloc] );
   copyblock( &globals.W[j%2]->matrix[iloc*globals.A->blksz], &globals.A->matrix[ijl*globals.A->blksz], globals.A->blksz );
}


/**
 * Kernel SYRK.
 * Updates A <-- A - WW' (aka SYRK) from iteration j. Block A is AL(ijl) which is A(iloc,jloc).
 * \param[in] j The iteration counter.
 * \param[in] iloc The global row index of the A block.
 * \param[in] jloc The global column index of the A block.
 * \param[in] ijl The DSBP block index of the A block.
 */
void kernel_SYRK( int j, int iloc, int jloc, int ijl )
{
   double ONE = 1.0;
   double MONE = -1.0;
   int N, K;

   N = min(globals.A->nb, globals.A->N - globals.A->nb * (globals.myrow + iloc*globals.nprow));
   K = globals.A->nb;

   if( opt_verbose ) {
      log_printf("SYRK(%d, %d, %d, %d)\n", j, iloc, jloc, ijl);
   }
   dsyrk_( "L", "N", &N, &K, &MONE, &globals.W[j%2]->matrix[globals.A->blksz * iloc], &globals.A->nb, &ONE, &globals.A->matrix[globals.A->blksz * ijl], &globals.A->nb );
}


/**
 * Kernel GEMM.
 * Updates A <-- A - WS' (aka GEMM) from iteration j. Block A is AL(ijl) which is A(iloc,jloc).
 * \param[in] j The iteration counter.
 * \param[in] iloc The global row index of the A block.
 * \param[in] jloc The global column index of the A block.
 * \param[in] ijl The DSBP block index of the A block.
 */
void kernel_GEMM( int j, int iloc, int jloc, int ijl )
{
   double ONE = 1.0;
   double MONE = -1.0;
   int M, N, K;

   M = min(globals.A->nb, globals.A->N - globals.A->nb * (globals.myrow + iloc*globals.nprow));
   N = min(globals.A->nb, globals.A->N - globals.A->nb * (globals.mycol + jloc*globals.npcol));
   K = globals.A->nb;

   if( opt_verbose ) {
      log_printf("GEMM(%d, %d, %d, %d)\n", j, iloc, jloc, ijl);
   }
   dgemm_( "N", "T", &globals.A->nb, &globals.A->nb, &globals.A->nb, &MONE, &globals.W[j%2]->matrix[globals.A->blksz * iloc], &globals.A->nb, &globals.S[j%2]->matrix[globals.A->blksz * jloc], &globals.A->nb, &ONE, &globals.A->matrix[globals.A->blksz * ijl], &globals.A->nb );
}


