#include <assert.h>
#include "static.h"
#include "kernels.h"
#include "comm.h"
#include "globals.h"
#include "utils.h"


/**
 * Static DSBP Cholesky algorithm. Uses depth-one lookahead, asynchronous communication algorithms, block data layout.
 * \param[in,out] A The DSBP matrix to factorize.
 * \param[in] nprow The number of processor rows.
 * \param[in] npcol The number of processor columns.
 * \param[in] myrow My process row index.
 * \param[in] mycol My process column index.
 * \return 0: OK, -1: definitely a bug in the software, 1: input matrix likely not SPD.
 */
int DSBPCholesky(DSBPMatrix *A, int nprow, int npcol, int myrow, int mycol)
{
   int Nb, blksz;
   int i, j, k;
   int il, jl, kl, ijl, is, ie, ks, ke;
   int pj, qj;
   int N, nb;
   int INFO;

   gINFO = 0;

   N = A->N;
   nb = A->nb;
   Nb = iceil(N, nb);
   blksz = nb * nb;

   globals.A = A;
   globals.nprow = nprow;
   globals.npcol = npcol;
   globals.myrow = myrow;
   globals.mycol = mycol;
   globals.R    = allocate_block_vector_buffer( 1, blksz );
   globals.W[0] = allocate_block_vector_buffer( A->rblks, blksz );
   globals.W[1] = allocate_block_vector_buffer( A->rblks, blksz );
   globals.S[0] = allocate_block_vector_buffer( A->cblks, blksz );
   globals.S[1] = allocate_block_vector_buffer( A->cblks, blksz );

   /* Init communication subsystem */
   comm_init( 1 + 2*A->rblks + 2*A->cblks, nprow, npcol, myrow, mycol );

   /**
    * Preamble (factorize first panel, j=0)
    */
   j = 0;
   
   /* Process (pj, qj) holds diagonal block A(j,j) */
   pj = j % nprow;
   qj = j % npcol;

   if( opt_verbose ) {
      log_printf("--- PROLOGUE ---\n");
   }

   if( mycol == qj ) {
      /**
       * Purpose:
       * This process is potentially involved in panel factorization j. 
       * This compound statement fulfills this process' role in that panel factorization.
       * It involves POTRF of diagonal block, TRSM of off-diagonal blocks and communication of R and W.
       */

      /**
       * Indices:
       * jl - Local column index of A(*,j)
       * il - First local row index (potentially out-of-bounds) in lower triangular part of A(*,j)
       * ijl - DSBP block index of local block AL(il,jl)
       */
      jl = G2Lmin( mycol, npcol, j );
      il = G2Lmin( myrow, nprow, j );
      ijl = A->cp[jl] - (A->ilastp - il);

      if( myrow == pj ) {
         /* Perform POTRF on diagonal block A(j,j). */
         kernel_POTRF( j, il, jl, ijl );
         /* Start to replicate the R-buffer. */
         startCommunicateR( j );
         
         il++;
         ijl++;
      } else if( G2Gmin( myrow, nprow, j+1 ) < Nb ) {
         /* This process is off-diagonal but holds at least one block in A(*,j) so it starts to replicate the R-buffer. */
         startCommunicateR( j );
      } 

      /**
       * Indices:
       * ie - Last local row index in A(*,j)
       * il - Ranges from top to bottom on non-diagonal local row indicdes of A(*,j)
       * ijl - DSBP block index corresponding with local block AL(il,jl)
       */
      ie = A->ilastp;
      if( il <= ie ) {
         /* Will enter loop below so we must wait for the replication of R. */
         comm_wait( &globals.R->bbs[0], j );
      }
      for( ; il <= ie; il++, ijl++ ) {
         /* Perform current TRSM */
         kernel_TRSM( j, il, jl, ijl );
         /* Start to replicate current W(il) */
         startCommunicateW( j, il );
         comm_poll( 0 );
      }
   } else {
      /**
       * Purpose:
       * This process does not take part in panel factorization j.
       * Therefore, its role is simply to start its replication of W.
       */
   
      /**
       * Indices:
       * jl - First local column index of trailing matrix from iteration j (which starts at global column j+1)
       * is - First local row index in lower triangular part of local column jl
       * ie - Last local row index of local column jl
       * il - Ranges between is and ie (which represents the set of all needed W blocks)
       */
      jl = G2Lmin( mycol, npcol, j+1 );
      if( jl < A->cblks ) {
         is = G2Lmin( myrow, nprow, mycol + jl*npcol );
         ie = A->ilastp;
         for( il = is; il <= ie; il++ ) {
            /* Start to replicate current W(il) */
            startCommunicateW( j, il );
         }
      }
   }

   /**
    * Purpose: 
    * To start all replications of non-root S-blocks since these are not initiated by a corresponding W-replication.
    */

   /**
    * Indices:
    * ks - First local column index of trailing matrix from iteration j (which starts at global column j+1)
    * ke - Last local column index
    * kl - Ranges from ks to ke (which represents the set of all potentially needed S blocks)
    */
   ks = G2Lmin( mycol, npcol, j+1 );
   ke = A->cblks - 1;
   for( kl = ks; kl <= ke; kl++ ) {
      k = mycol + kl*npcol;
      /* Check that (1) we are not root for S-replication on column A(*,k), and (2) we have at least one lower triangular block on column A(*,k) */
      if( myrow != k % nprow && G2Gmin( myrow, nprow, k ) < Nb ) {
         startCommunicateS( j, kl );
      }
   }

   if( opt_verbose ) {
      log_printf("--- LOOP ---\n");
   }
   
   /**
    * Purpose:
    * Let j range over remaining panels and complete the factorization of A.
    */
   for( j = 1; j < Nb; j++ ) {
      /* Process (pj, qj) holds diagonal block A(j,j) */
      pj = j % nprow;
      qj = j % npcol;

      if( mycol == qj ) {
         /**
          * Purpose:
          * This process is potentially involved in panel factorization j.
          * This compound statement fulfills this process' role in that panel factorization.
          * It involves an interleaving of delayed updates and panel factorization.
          */

         /**
          * Indices:
          * jl - Local column index of A(*,j)
          * il - First local row index of lower triangular part of A(*,j)
          * ijl - DSBP block index of local block AL(il,jl)
          */
         jl = G2Lmin( mycol, npcol, j );
         il = G2Lmin( myrow, nprow, j );
         ijl = A->cp[jl] - (A->ilastp - il);
         if( myrow == pj ) {
            /* Wait for W(il) from previous iteration */
            comm_wait( &globals.W[(j-1)%2]->bbs[il], j-1 ); 
            /* Perform SYRK from previous iteration */
            kernel_SYRK( j-1, il, jl, ijl );
            /* Perform current POTRF */
            kernel_POTRF( j, il, jl, ijl );
            /* Start to replicate the R-buffer */
            startCommunicateR( j );
            comm_poll( 0 );

            il++;
            ijl++;
         } else if( G2Gmin( myrow, nprow, j+1 ) < Nb ) {
            /* This process has at least one block of lower triangular part of A(*,j) and hence needs R. */
            startCommunicateR( j );
         }

         /**
          * Indices:
          * ie - Last local row index of column A(*,j)
          * il - Ranges over off-diagonal local row indices on column A(*,j)
          */
         ie = A->ilastp;
         if( il <= ie ) {
            /* Will enter loop below so wait for R. */
            comm_wait( &globals.R->bbs[0], j ); 
            /* Wait for S(jl) from previous iteration */
            comm_wait( &globals.S[(j-1)%2]->bbs[jl], j-1 ); 
         }
         for( ; il <= ie; il++, ijl++ ) {
            /* Wait for W(il) from previous iteration */
            comm_wait( &globals.W[(j-1)%2]->bbs[il], j-1 );
            /* Perform previous GEMM */
            kernel_GEMM( j-1, il, jl, ijl );
            comm_poll( 0 );
            /* Perform current TRSM */
            kernel_TRSM( j, il, jl, ijl );
            /* Start to replicate current W(il) */
            startCommunicateW( j, il );
            comm_poll( 0 );
         }
      } else {
         /**
          * Purpose:
          * This process does not take part in panel factorization j.
          * Therefore, it waits for previous W replications and starts current W replications.
          */

         /**
          * Indices:
          * jl - First local column index of trailing matrix from previous iteration (which starts at global column j=(j-1)+1).
          * is - First local row index of local column AL(*,jl)
          * ie - Last local row index
          * il - Ranges from is to ie (which represents the set of needed W-blocks)
          */
         jl = G2Lmin( mycol, npcol, j );
         if( jl < A->cblks ) {
            is = G2Lmin( myrow, nprow, mycol + jl*npcol );
            ie = A->ilastp;
            for( il = is; il <= ie; il++ ) {
               /* Wait for previous W(il) */
               comm_wait( &globals.W[(j-1)%2]->bbs[il], j-1 ); 
            }
         }

         /**
          * Indices:
          * jl - First local column index of trailing matrix from current iteration (which starts at global column j+1).
          * is - First local row index of local column AL(*,jl)
          * ie - Last local row index
          * il - Ranges from is to ie (which represents the set of needed W-blocks)
          */
         jl = G2Lmin( mycol, npcol, j+1);
         if( jl < A->cblks ) {
            is = G2Lmin( myrow, nprow, mycol + jl*npcol );
            ie = A->ilastp;
            for( il = is; il <= ie; il++ ) {
               startCommunicateW( j, il );
            }
         }
      }

      /**
       * Purpose:
       * Regardless if process is involved in panel factorization j or not, it starts replication of S-blocks for the current iteration.
       * Only replication of blocks for which this proces is a non-root get started here (the others implicitly by a W-replication).
       */

      /**
       * Indices:
       * ks - First local column index of trailing matrix from current iteration (which starts at global column j+1).
       * ke - Last local column index
       * kl - Ranges from ks to ke (which represents the set of potentially needed S-blocks)
       */
      ks = G2Lmin( mycol, npcol, j+1 );
      ke = A->cblks - 1;
      for( kl = ks; kl <= ke; kl++ ) {
         k = mycol + kl*npcol;
         /* Check that (1) this process is not the root, and (2) it has at least one lower triangular block of column A(*,k) */
         if( myrow != k % nprow && G2Gmin( myrow, nprow, k ) < Nb ) {
            /* Start to replicate current S(kl) */
            startCommunicateS( j, kl );
         }
      }

      /**
       * Purpose: 
       * Complete the updates from previous iteration.
       * Recall that column A(*,j) has already been updated so the remaining trailing part starts at column j+1.
       */

      /**
       * Indices:
       * ks - First local column index of non-updated trailing matrix from previous iteration (which starts at column j+1=(j-1)+2).
       * ke - Last local column index.
       * kl - Ranges from ks to ke.
       * ie - Last local row index.
       * il - Ranges from the first local row index on lower triangular part of column kl to ie.
       * ijl - DSBP block index of local block A(il,kl).
       * i - Global row index corresponding with il.
       * k - Global column index corresponding with kl.
       */
      ks = G2Lmin( mycol, npcol, j+1 );
      ke = A->cblks - 1;
      ie = A->ilastp;
      for( kl = ks; kl <= ke; kl++ ) {
         k = mycol + kl*npcol;
         il = G2Lmin( myrow, nprow, k );
         i = myrow + il*nprow;
         ijl = A->cp[kl] - (A->ilastp - il);
         if( il <= ie ) {
            /* Wait for previous S(kl) */
            comm_wait( &globals.S[(j-1)%2]->bbs[kl], j-1 ); 
         }
         if( i == k ) {
            /* Perform previous SYRK */
            kernel_SYRK( j-1, il, kl, ijl );
            comm_poll( 0 );
            
            il++;
            ijl++;
         }
         for( ; il <= ie; il++, ijl++ ) {
            /* Perform previous GEMM */
            kernel_GEMM( j-1, il, kl, ijl );
            comm_poll( 0 );
         }
      }
   }

   if( opt_verbose ) {
      log_printf("--- EPILOGUE ---\n");
   }

   /* Wait for all communication to finish */
   comm_wait_all( );

   /* Deallocate communication data */
   comm_finalize( );

   /* Deallocate buffers */
   deallocate_block_vector_buffer( globals.R );
   deallocate_block_vector_buffer( globals.W[0] );
   deallocate_block_vector_buffer( globals.W[1] );
   deallocate_block_vector_buffer( globals.S[0] );
   deallocate_block_vector_buffer( globals.S[1] );

   /* Find if error detected */
   MPI_Allreduce(&gINFO, &INFO, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
   if( INFO < 0 ) {
      return -1;
   } else {
      MPI_Allreduce(&gINFO, &INFO, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
      if( INFO > 0 ) {
         return 1;
      } else {
         return 0;
      }
   }
}

