#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include "test.h"
#include "globals.h"


extern int getopt(int argc, char * const argv[], const char *optstring);
extern char *optarg;


/**
 * Compute number of rows/columns mapped to a process row/column.
 * \param[in] me The process ID (0..np-1)
 * \param[in] np The number of processes.
 * \param[in] N The number of elements.
 * \param[in] nb The block size.
 * \return The number of rows/columns mapped to me.
 */
int numroc( int me, int np, int N, int nb )
{
   int num;

   num = ((N / nb) / np) * nb;
   if( me < (N / nb) % np )
      num += nb;
   else if( me == (N / nb) % np )
      num += N % nb;

   return num;
}



/**
 * main
 * Usage: mpirun -np np ./prog -r Pr -c Pc -N N -b nb [-v]
 * Purpose:
 * This routine is intended to show how to interface with the various routines. 
 */
int main(int argc, char *argv[])
{
   int np, me, myrow, mycol, nprow, npcol;
   FILE *out;
   char fnout[100];
   int c;
   int N, nb;
   DSBPMatrix A;
   double *x;
   int Nr, Nc, ldB;
   int i, j, ig, jg;
   double *B;
   int ret;
   double Tp;
   double GFLOPS;

   srand(time(NULL));

   MPI_Init(&argc, &argv);

   MPI_Comm_size(MPI_COMM_WORLD, &np);
   MPI_Comm_rank(MPI_COMM_WORLD, &me);

   /**
    * Parse options.
    */
   while( (c = getopt(argc, argv, "r:c:N:b:v")) != -1 ) {
      switch( c ) {
         case 'r':
            nprow = atoi(optarg);
            break;

         case 'c':
            npcol = atoi(optarg);
            break;

         case 'N':
            N = atoi(optarg);
            break;

         case 'b':
            nb = atoi(optarg);
            break;

         case 'v':
            opt_verbose = 1;
            break;

         default:
            fprintf(stderr, "Unknown option: %c\n", c);
            MPI_Abort(MPI_COMM_WORLD, -1);
            break;
      }
   }

   /**
    * Check options.
    */
   if( nprow < 1 || npcol < 1 || nprow*npcol != np ) {
      fprintf(stderr, "Inconsistent mesh: %d x %d != %d\n", nprow, npcol, np);
      MPI_Abort(MPI_COMM_WORLD, -1);
   }
   if( N < 1 || (opt_verbose && (N > 500 || nb > 500)) ) {
      fprintf(stderr, "Unallowed matrix size: %d (too small or too large when verbose is on)\n", N);
      MPI_Abort(MPI_COMM_WORLD, -1);
   }

   /**
    * Begin main body.
    */
   mycol = me % npcol;
   myrow = me / npcol;

   if( opt_verbose ) {
      sprintf(fnout, "out_%d_%d.m", myrow, mycol);
      out = fopen(fnout, "w");
      assert(out != NULL);
      log_init(nprow, npcol, myrow, mycol);
      log_printf("--- START ---\n");
   }

   /**
    * Local matrix is Nr x Nc.
    */
   Nr = numroc( myrow, nprow, N, nb );
   Nc = numroc( mycol, npcol, N, nb );
   ldB = Nr;
   
   /**
    * Generate the SPD matrix B = I + x*x' where ||x|| = 1.
    */
   B = abortmalloc( sizeof(double) * Nr * Nc );
   x = abortmalloc( sizeof(double) * N );
   if( myrow + mycol == 0 ) {
      /**
       * Generate x such that ||x|| = 1.
       */
      double nrm = 0.;

      for( i = 0; i < N; i++ ) {
         x[i] = (double) rand() / RAND_MAX;
         nrm += x[i]*x[i];
      }
      nrm = 1. / sqrt(nrm);
      for( i = 0; i < N; i++ ) {
         x[i] = x[i] * nrm;
      }
      if( opt_verbose ) {
         /**
          * Print x in MATLAB readable form.
          */
         fprintf( out, "x = [" );
         for( i = 0; i < N; i++ ) {
            fprintf( out, "%.16e ", x[i] );
         }
         fprintf( out, "];\n" );
      }
   }
   MPI_Bcast(x, N, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   /**
    * Generate B by computing local part of B = I + x*x'.
    */
   for( j = 0; j < Nc; j++ ) {
      for( i = 0; i < Nr; i++ ) {
         ig = (myrow + (i/nb)*nprow)*nb + i%nb;
         jg = (mycol + (j/nb)*npcol)*nb + j%nb;
         if( ig == jg )
            B[i + j*ldB] = 1;
         else
            B[i + j*ldB] = 0;
         B[i + j*ldB] += x[ig]*x[jg];
      }
   }

   /**
    * Convert B into DSBP A.
    */
   DSBPFromFull( &A, B, ldB, nprow, npcol, myrow, mycol, N, nb );

   /**
    * Compute the Cholesky factorization of DSBP A.
    */
   MPI_Barrier(MPI_COMM_WORLD);
   Tp = get_timestamp();
   ret = DSBPCholesky( &A, nprow, npcol, myrow, mycol );
   MPI_Barrier(MPI_COMM_WORLD);
   Tp = get_timestamp() - Tp;

   if( opt_verbose ) {
      /**
       * Print DSBP A in MATLAB readable form.
       */
      DSBP2Matlab( out, &A, nprow, npcol, myrow, mycol );
   }

   /**
    * Free A and B.
    */
   DSBPDestroy( &A );
   free( B );
   
   if( opt_verbose ) {
      fclose(out);
      log_printf("--- END ---\n");
      log_finalize();
   }

   if( me == 0 ) {
      GFLOPS = ((double)N*(double)N*(double)N/3.*1E-9) / Tp;
      printf("Results\n"
             "------------------------------------------------------------\n"
             "Mesh Size:      %2d x %2d\n"
             "Matrix Size:     N = %d\n"
             "Block Size:     nb = %d\n"
             "Execution Time: Tp = %.6lf seconds\n"
             "Performance:     P = %.3lf GFLOP/s\n"
             "Efficiency:   Pavg = %.3lf GFLOP/s per process\n"
             "Return Value:        %d (0: OK, -1: Error, 1: Input not SPD)\n"
             "\n"
             , nprow, npcol, N, nb, Tp, GFLOPS, GFLOPS/np, ret);
   }
   
   MPI_Finalize();

   return 0;
}
