#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"


/**
 * Copies 'blksz' contiguous elements from src to dst.
 * \param[out] dst Destination.
 * \param[in] src Source.
 * \param[in] blksz Number of contiguous elements to copy.
 */
void copyblock( double *dst, double *src, int blksz )
{
   int i;

   for( i = 0; i < blksz; i++ ) {
      dst[i] = src[i];
   }
}


/**
 * Allocates 'bytes' bytes of memory using malloc but calls MPI_Abort if out of memory.
 * \param[in] bytes The number of bytes to allocate.
 * \return A non-NULL pointer to allocated memory.
 */
void* abortmalloc( size_t bytes )
{
   void *data;

   if( ((int) bytes) < 0 ) {
      fprintf( stderr, "Tried to allocate too much memory - overflow\n" );
      fflush( stderr );
      MPI_Abort( MPI_COMM_WORLD, -1 );
   }

   /* Try to allocate */
   data = malloc( bytes );

   /* MPI_Abort if failed to allocate */
   if( data == NULL ) {
      fprintf( stderr, "!!! Out of memory\n" );
      fflush( stderr );
      MPI_Abort( MPI_COMM_WORLD, 1 );
   }
   
   return data;
}


/**
 * When run in verbose mode (-v) the program outputs a set of log files (one per process) containing timestamped event details.
 * The data and functions below are used to implement this feature.
 */
static FILE *fp = NULL;
static double prev_ts = 0; /* Previous timestamp (used to calculate duration since last event) */
static double base = 0; /* Time base (to get shorter time printings) */


/**
 * Returns the current time in seconds since some arbitrary but constant time base.
 * \return Time in seconds.
 */
double get_timestamp(void)
{
   struct timeval tv;

   gettimeofday(&tv, NULL);

   return tv.tv_sec + 1E-6*tv.tv_usec;
}


/**
 * Init the log by opening the log file and determine a time base.
 * \param[in] nprow The number of process rows.
 * \param[in] npcol The number of process columns.
 * \param[in] myrow My process row index.
 * \param[in] mycol My process column index.
 */
void log_init(int nprow, int npcol, int myrow, int mycol)
{
   char fn[100];

   base = floor(get_timestamp());
   
   sprintf(fn, "log_%d_%d.txt", myrow, mycol);
   fp = fopen(fn, "w");
   assert(fp != NULL);
   fprintf(fp, "TimeBase: %.6f\n", base);
}


/**
 * Timestamped printf(.) but to the log file instead of stdout.
 */
void log_printf(char *format, ...)
{
   va_list ap;
   double ts;

   va_start(ap, format);
   ts = get_timestamp() - base;
   if( prev_ts == 0 )
      fprintf(fp, "[%.6f] ", ts);
   else 
      fprintf(fp, "[%.6f] (+%.0f) ", ts, 1E6*(ts - prev_ts));
   prev_ts = ts;
   vfprintf(fp, format, ap);
   fflush(fp);
   va_end(ap);
}


/**
 * Closes the log file.
 */
void log_finalize(void)
{
   fclose(fp);
}
