#ifndef UTILS_H
#define UTILS_H

#include <stdlib.h>
#include <assert.h>


void* abortmalloc( size_t bytes );
void copyblock( double *dst, double *src, int blksz );
double get_timestamp(void);



/**
 * Computes ceil(a / b) for non-negative numbers.
 * \param[in] a Numerator.
 * \param[in] b Denominator.
 * \return ceil(a / b)
 */
static int iceil( int a, int b )
{
   return (a + b - 1) / b;
}


/**
 * Computes floor(a / b) for positive b and semi-positive a (-b <= a).
 * \param[in] a The numerator.
 * \param[in] b The denominator.
 * \return floor(a / b)
 */
static int ifloor( int a, int b )
{
   return (a + b) / b - 1;
}


/**
 * Finds the minimum k such that (me + k*np) >= roc and return k.
 * This effectively finds the first local index mapped to me in the submatrix starting at block roc.
 * \param[in] me My process index.
 * \param[in] np Number of processes.
 * \param[in] roc RHS.
 * \return k.
 */
static int G2Lmin( int me, int np, int roc ) 
{
   return iceil( roc - me, np );
}


/**
 * Finds the minimum k such that (me + k*np) >= roc and return the LHS.
 * This effectively finds the first global index mapped to me in the submatrix starting at block roc.
 * \param[in] me My process index.
 * \param[in] np Number of processes.
 * \param[in] roc RHS.
 * \return me + k*np.
 */
static int G2Gmin( int me, int np, int roc ) 
{
   return me + G2Lmin( me, np, roc ) * np;
}


/**
 * Finds the maximum k such that (me + k*np) <= roc and return k.
 * This effectively finds the last local index mapped to me in the submatrix ending at block roc.
 * \param[in] me My process index.
 * \param[in] np Number of processes.
 * \param[in] roc RHS.
 * \return k.
 */
static int G2Lmax( int me, int np, int roc )
{
   return ifloor( roc - me, np );
}


/**
 * Finds the maximum k such that (me + k*np) <= roc and return the LHS.
 * This effectively finds the last global index mapped to me in the submatrix ending at block roc.
 * \param[in] me My process index.
 * \param[in] np Number of processes.
 * \param[in] roc RHS.
 * \return me + k*np.
 */
static int G2Gmax( int me, int np, int roc )
{
   return me + G2Lmax( me, np, roc ) * np;
}


static int min( int a, int b )
{
   return (a < b ? a : b);
}


static int max( int a, int b )
{
   return (a > b ? a : b);
}


void log_init(int nprow, int npcol, int myrow, int mycol);
void log_printf(char *format, ...);
void log_finalize(void);

#endif
